(*  Title:      HOL/TPTP/Satallax_Reconstruct.thy
    Author:     Nik Sultana, Cambridge University Computer Laboratory

Reconstruction of TPTP-encoded proofs produced by Satallax.
*)

theory Satallax_Reconstruct
imports
  Satallax
  TPTP_Proof_Reconstruction
begin

section "Satallax proof reconstruction"

(*FIXME
  - The parser reverses the list of parents in AFs. At some point this should be fixed, but care must
    be taken to update all dependent code. cf formula_meaning.
  - Use "val"s for strings denoting names of inference rules and lemmas, rather than having constants strewn around.
*)


section "Configuration"

ML {*
  val debug_satallax_reconstruction = Attrib.setup_config_bool @{binding debug_satallax_reconstruction} (K false)
  fun debug_mode ctxt = Config.get ctxt debug_satallax_reconstruction
*}


section "Constants"

ML {*
(*Suffix for names of definitions*)
val def_suffix = "_def"

(*Suffix for names of skolem equations*)
val skolemeq_suffix = "_skolemeq"

(*Our marker for the terminal node during the
  traversal (of the proof graph) that generates
  the cut program/plan.*)
val terminal_node = "-!TERM"

(*We start reconstructing a Satallax proof from the node with this label*)
val starting_nodeid_of_proof = "1"

(*We start reconstructing a Satallax proof from a hypothesis having this label*)
val starting_assumption_of_proof = "h0"

(*Constant and equation used when Skolemisation is applied to a quantified variable that doesn't
  occur free inside the formula body -- we don't need a fresh Skolem constant for that variable,
  since the Skolem constant won't appear anywhere.*)
val demmy = @{term "True"}
val demmy_eq =
  (demmy, demmy)
  |> HOLogic.mk_eq
  |> HOLogic.mk_Trueprop
*}


section "General"

ML {*
fun term_to_string ctxt t =
  Print_Mode.with_modes [""]
    (fn () => Output.output (Syntax.string_of_term ctxt t)) ()
*}

ML {*
(*FIXME move to library*)
fun fold_opt f opt x =
  case opt of
    NONE => x
  | SOME y => f y x
*}


section "Term manipulation"

ML {*
(*Change a formula (~ ! x. P x) into a term
  (SOME x. ~ P x)
  This is used when handling Skolemisation steps,
  to generate Skolem equations that define Skolem constants.
*)
fun negall_to_witness t =
  case t of
    (negation as Const (@{const_name HOL.Not}, _)) $
            (Const (@{const_name HOL.All}, _) $
              Abs (var, ty, body)) =>
             (Const (@{const_name Hilbert_Choice.Eps}, (ty --> HOLogic.boolT) --> ty) $
              Abs (var, ty, negation $ body))
  | _ => error "negall_to_witness applied to wrong formula shape"
*}

ML {*
(*Change a formula Ex P into a term
  (SOME x. P x)
  This is used when handling Skolemisation steps,
  to generate Skolem equations that define Skolem constants.
*)
fun ex_to_witness t =
  case t of
    (Const (@{const_name Ex}, ty) $ t) =>
     let
       val witness_ty =
         dest_funT ty
         |> fst
         |> dest_funT
         |> fst
      in
        (Const (@{const_name Hilbert_Choice.Eps}, (witness_ty --> HOLogic.boolT) --> witness_ty) $
          Abs ("", witness_ty, t $ Bound 0))
      end
  | _ => error "ex_to_witness applied to wrong formula shape"
*}


subsection "Inference-rule generation"

text "The functions in this section are used to generate terms like the ones preceding each function."

ML {*
@{term "\<And>R. H0 \<Longrightarrow> (\<lbrakk>H0; H1\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"};
(*FIXME for better style use @{const_name} instead of strings.*)
(*if Ha is NONE then we bind a metavariable, otherwise we insert the
  constant of which name it carries*)
fun mk_elim_inference1 Ha Hb =
let
  val (atom, f_term) =
    case Ha of
      NONE => (Bound 1, fn t =>
      Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
             Abs ("Q", @{typ "bool"}, t))
    | SOME Ha' => (Const (Ha', @{typ "bool"}), fn t => t)
  val body =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ atom) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ atom) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))
in
  f_term body
end;

(*NOTE that H0 isn't inserted in the intermediate steps*)
@{term "\<And>R. H0 \<Longrightarrow> (H1 \<Longrightarrow> R) \<Longrightarrow> (H2 \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_elim_inference2 Ha Hb Hc =
  Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))));

@{term "\<And>R. H0 \<Longrightarrow> (\<lbrakk>H0; H1; H2\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_elim_inference1_2 Ha Hb Hc =
  Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)));

@{term "Ha \<Longrightarrow> Hb \<Longrightarrow> False"};
fun mk_conflict Ha Hb =
  Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
           (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
             (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const ("HOL.False", @{typ "bool"})));

@{term "\<And>R. H0 \<Longrightarrow> (\<lbrakk>H0; H1; H2\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> (\<lbrakk>H0; H3; H4\<rbrakk> \<Longrightarrow> R) \<Longrightarrow>R"};
fun mk_elim_inference2_5 Ha Hb Hc Hd He =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hd, @{typ "bool"})) $
                       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (He, @{typ "bool"})) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))));

@{term "\<And>R. \<lbrakk>H0; H1\<rbrakk> \<Longrightarrow> (\<lbrakk>H0; H1; H2\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_elim_inference2_1 Ha Hb Hc =
Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
  Abs ("R", @{typ "bool"},
    Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
      (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
        (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
          (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
            (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
            (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
              (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
              (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
          (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))));

@{term "\<And>R. \<lbrakk>H0; H1\<rbrakk> \<Longrightarrow> (\<lbrakk>H0; H1; H2\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> (\<lbrakk>H0; H1; H3\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_elim_inference2_2 Ha Hb Hc Hd =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
     Abs ("R", @{typ "bool"},
       Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
         (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
           (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
             (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
               (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
             (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hd, @{typ "bool"})) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
               (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))));

@{term "\<And>R. \<lbrakk>H0; H1\<rbrakk> \<Longrightarrow> (\<lbrakk>H0; H1; H2; H3\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> (\<lbrakk>H0; H1; H4; H5\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_elim_inference2_6 Ha Hb Hc Hd He Hf =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"} ) $ Const (Hc, @{typ "bool"})) $
                         (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hd, @{typ "bool"})) $
                           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))))) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                         (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (He, @{typ "bool"})) $
                           (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                             (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hf, @{typ "bool"})) $
                             (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))))) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))));

(*based on TCut in Satallax.thy, since that rule will be used
  to validate instances of the terms generated by mk_cut*)
@{term   "\<And> P R. P
   \<Longrightarrow> (\<lbrakk>P; H0\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>P; H1\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"};
fun mk_cut Ha Hb =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("P", @{typ "bool"},
             Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
               Abs ("R", @{typ "bool"},
                 Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 1) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 1) $
                       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                         (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 1) $
                         (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $
                             (Const (Hb, @{typ "bool"}))) $
                           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))));

@{term "\<And> P R. H0 \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_exfalso Ha =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("P", @{typ "bool"},
             Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
               Abs ("R", @{typ "bool"},
                 Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
                   (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                     (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 1) $
                       (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)) $
                     (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))));

(*like mk_elim_inference1 but forgets about H0*)
@{term "\<And>R. H0 \<Longrightarrow> (H1 \<Longrightarrow> R) \<Longrightarrow> R"};
fun arbitrary_elim Ha Hb =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
           Abs ("R", @{typ "bool"},
             Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
                   (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)));

(*used by rewriting tactics, to rewrite H0 to H1.
  see tab_rew in Satallax.thy*)
@{term "\<And>R. H0 \<Longrightarrow> (H1 \<Longrightarrow> R) \<Longrightarrow> R"};

@{term "\<And>R. H0 \<Longrightarrow> H1 \<Longrightarrow> (H0 \<Longrightarrow> H1 \<Longrightarrow> H2 \<Longrightarrow> R) \<Longrightarrow> R"};
fun mk_tabmat_3 Ha Hb Hc =
 Const ("Pure.all", @{typ "(bool \<Rightarrow> prop) \<Rightarrow> prop"}) $
   Abs ("R", @{typ "bool"},
     Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
       (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $ (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
         (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
           (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
             (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Ha, @{typ "bool"})) $
             (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
               (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hb, @{typ "bool"})) $
               (Const ("Pure.imp", @{typ "prop \<Rightarrow> prop \<Rightarrow> prop"}) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Const (Hc, @{typ "bool"})) $
                 (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0)))) $
           (Const ("HOL.Trueprop", @{typ "bool \<Rightarrow> prop"}) $ Bound 0))));
*}


section "Proof querying"

ML {*
fun is_known_lemma node_id =
  List.exists (fn lem_name => node_id = lem_name)
  ["eq_ind_r", "eq_ind", "eq_sym"]

fun is_assumption_node fms node_id =
  (*Consider known lemmas to be assumptions for the purpose
    of this classification. They are more than assumptions,
    of course, since they are lemmas, but they can be distinguished
    by using the is_known_lemma predicate.*)
  is_known_lemma node_id orelse
  TPTP_Reconstruct.node_info fms (fn info : TPTP_Reconstruct.formula_meaning_body =>
    ((*warning (node_id ^ " :: " ^ PolyML.makestring (#source_inf_opt info));*)
     #role info = TPTP_Syntax.Role_Assumption
     orelse #role info = TPTP_Syntax.Role_Negated_Conjecture)) node_id
*}


section "Preprocessing"

ML {*
(*Remove the (usually singular) AFs having a "theorem" role.*)
fun remove_theorem_roled_AFs (pannot : TPTP_Reconstruct.proof_annotation) =
  {problem_name = #problem_name pannot,
   skolem_defs = #skolem_defs pannot,
   defs = #defs pannot,
   axs = #axs pannot,
   meta =
     #meta pannot
     |> filter (fn (name, {role, ...}) =>
          if role = TPTP_Syntax.Role_Theorem then
            (*NOTE Satallax usually names the theorem node "0".*)
            (@{assert} (name = "0");
             false)
          else true)
  }
*}

ML {*
(*List the assumptions in the proof.
  These can be recognised for having labels of the
  form "hD+" where "h" is the character "h", and
  "D+" is one or more decimal numerals.
  Rather than use the labels, a more reliable way
  of picking assumptions out involves looking at
  their role: these are either "assumption" or
  "negated_conjecture".*)
fun extract_assumptions
  (pannot : TPTP_Reconstruct.proof_annotation) : TPTP_Reconstruct.formula_meaning list =
let
  fun extract role =
    #meta pannot
    |> TPTP_Reconstruct.filter_by_role role
in
  extract TPTP_Syntax.Role_Assumption
  @ extract TPTP_Syntax.Role_Negated_Conjecture
end
*}


section "(Conservative) theory extension"

ML {*
(*NOTE this code is based on absorb_skolem_def*)
fun absorb_assumption_def const_name conclusion (pannot : TPTP_Reconstruct.proof_annotation) thy =
(@{assert} (const_name <> "");
let
  (*Assumption formulas are given names. These names
    name constants that are defined to be equivalent
    to the formulas they allude to; thus a name
    abbreviates an assumption formula.*)
  val bnd =
      Long_Name.base_name const_name
      |> Binding.name
      |> Binding.qualify false
        ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)

  (*Add the constant to the theory*)
  val (const, thy') =
    Sign.declare_const_global
    ((bnd, HOLogic.boolT), NoSyn) thy

  val _ =
    if debug_mode (Context.Theory thy |> Context.proof_of) then
      warning ("Adding def:" ^ const_name)
    else ()

  val bnd_def =
    Binding.suffix_name def_suffix bnd

  val bnd_def_name =
    Binding.name_of bnd_def
    |> Long_Name.base_name

  val _ =
    if debug_mode (Context.Theory thy |> Context.proof_of) then
      warning ("Added def:" ^ bnd_def_name)
    else ()

in
  case try (Thm.axiom thy) bnd_def_name of
      SOME _ => thy
    | NONE =>
        let
          val t = Logic.mk_equals (const, HOLogic.dest_Trueprop conclusion)
          val ((_, thm), thy'') = Thm.add_axiom_global (bnd_def, t) thy'
          val (_, thy''') = Global_Theory.add_thm ((bnd_def, thm), []) thy''
        in
          thy'''
        end
end)
*}


subsection "Skolemisation"

ML {*
(*Intermediate data structure that is used when analysing
  the proof to generate Skolem equations*)
type skolem_data =
  {node_name : string,
   sk_term : term,
   sk_eq : term option,
   (*The parent containing the formula
     _before_ Skolemisation is applied*)
   parent1 : TPTP_Reconstruct.formula_meaning,
   (*The parent containing the formula 
     _after_ Skolemisation is applied*)
   parent2 : TPTP_Reconstruct.formula_meaning}

(*Extract Skolem equations from a proof, and add them
  to the theory.
  NOTE Rather oddly it seems that the arity of Skolem constants
       isn't an issue in this calculus. I think this might be
       because formulas are always ground -- unlike Leo2 clauses,
       which can have free variables.
*)
fun absorb_skolem_equations 
   (inference_name : string)
   (fmla_to_witness : term -> term)
   (pannot : TPTP_Reconstruct.proof_annotation) : theory -> theory =
  (*Get all Skolemisation steps*)
  nodes_by_inference (#meta pannot) inference_name
  (*Extend to include names of parents*)
  |> map (fn name : string => (name, TPTP_Reconstruct.parents_of_node (#meta pannot) name))
  (*Replace names of parents with the parent nodes' records*)
  |> map (apsnd (map (fn name : string => (name, TPTP_Reconstruct.node_info (#meta pannot) I name))))
  (*Remove parent nodes that aren't assumptions*)
  |> map (apsnd (filter (fn (_, {role, ...}) =>
      role = TPTP_Syntax.Role_Assumption orelse
       role = TPTP_Syntax.Role_Negated_Conjecture)))
  (*Type check*)
  |> (fn skoldatas : (string * TPTP_Reconstruct.formula_meaning list) list => skoldatas)
  (*Ensure that there are 2 parents*)
  |> (fn skoldatas =>
      (@{assert} (fold (fn (_, skoldata) => fn b =>
         b andalso length skoldata = 2) skoldatas true);
       skoldatas))
  (*Turn the list of pairs into a 2-tuple of parents, since we've established
    that there must be 2 of them*)
  |> (map (apsnd
    (fn [h1, h2] => ((h1, h2) : (TPTP_Reconstruct.formula_meaning * TPTP_Reconstruct.formula_meaning))
        | _ => error "Only 2 parents expected")))

  (*Calculate the difference in constants, to determine the Skolem constant*)
  |> map (fn (name, (fm1 as (_, info1), fm2 as (_, info2))) =>
       let
         val diff =
           consts_in (#fmla info2)
           |> list_diff (consts_in (#fmla info1))
       in
        {node_name = name,
         sk_term =
            (*Here we check whether we should use the _notfree variants of TNegAll and TEx.
              The logic down the line will be sensitive to a choice made at this point.
              If "null diff" it means that no Skolem constants were introduced into the term --
              and this happens if the existentially-bound variable wasn't free in the formula
              body. We detect this here, and use dummy Skolem constants to avoid creating new ones 
              -- these will never show up in the proof anyway. The only thing we will know about
              this constant is that it is equal to itself.
            *)
            if null diff then demmy
            else the_single diff (*there should only be one constant here*),
         sk_eq =
            (*As with setting sk_term, we set sk_eq with a trivial equation, to ensure that it won't
              get set down the line, and to signal to code later on to use the _notfree variants of
              TNegaAll and TEx.*)
            if null diff then
              SOME demmy_eq
            else NONE,
         parent1 = fm1,
         parent2 = fm2}
       end)
  (*Type check*)
  |> (fn skoldatas : skolem_data list => skoldatas)
  
  |> map (fn skol_data =>
    if is_some (#sk_eq skol_data) then skol_data
    else
    let
      (*Turn a formula into the definiens of a Skolem equation*)
      val definiens =
        (#fmla o snd o #parent2) skol_data
        |> HOLogic.dest_Trueprop
        |> fmla_to_witness
      (*Turn this into an equation*)
      val definition =
        HOLogic.mk_eq (#sk_term skol_data, definiens)
        |> HOLogic.mk_Trueprop
    in
        {node_name = #node_name skol_data,
         sk_term = #sk_term skol_data,
         sk_eq = SOME definition,
         parent1 = #parent1 skol_data,
         parent2 = #parent2 skol_data}
    end)
  (*Create a binding for it*)
  |> fold (fn skol_data => fn thy' =>
      let
        val binding =
          #node_name skol_data
          |> Binding.name
          |> Binding.suffix_name skolemeq_suffix
          |> Binding.qualify false
            ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)
  
        val ((s, thm), thy'') =
          Thm.add_axiom_global (binding, the (#sk_eq skol_data)) thy'

        val _ =
          if debug_mode (Context.Theory thy'' |> Context.proof_of) then
            warning ("Added skolem equation: " ^ s)
          else ()

        val (_, thy''') = Global_Theory.add_thm ((binding, thm), []) thy''
      in
        thy'''
      end)
*}


section "Proof transformation"

ML {*
(*Beginning at start_node_id, we work our way through the proof graph,
  transforming the formula at each node that has role "plain"
  (i.e., we ignore assumptions and other roles).
  In Satallax proofs, "plain" nodes carry the formula "False",
  which is the conclusion of the refutation.
  This proof transformation changes that formula into
  implications between assumptions -- specifically, the
  assumptions that are used and inferred in that node.

  To form this formula we use the parent annotation.
  There are 2 kinds of nodes in Satallax proofs:
  lemmas/assumptions, and inferences -- the former
  carry formula information, whereas the second carry
  inference information. In this proof transformation
  we use the former to modify the formula carried
  in the latter. This will have 2 effects:
   - the formula will encode the inference formula
     that will be emulated.
   - traversing inference nodes will generate the
     cut program -- i.e., we can ignore assumption
     nodes during the traversal.
*)
fun refine_conclusion (start_node_id : string) (pannot : TPTP_Reconstruct.proof_annotation) thy : TPTP_Reconstruct.proof_annotation =
let
  val fms = #meta pannot
  (*Traversal over inference nodes.*)
  fun traverse (to_examine : string list) (acc : TPTP_Reconstruct.formula_meaning list) =
    case to_examine of
      [] => acc
    | node_name :: rest =>
      (*We should never encounter assumption nodes while traversing.
        Such nodes are never put into to_examine.*)
      if (is_assumption_node fms node_name) then
        error "Cannot traverse to assumption"
      else
      let
        val node = TPTP_Reconstruct.node_info fms (fn x => x) node_name

        val _ =
          if debug_mode (Context.Theory thy |> Context.proof_of) then
            warning ("Traversing " ^ @{make_string} node)
          else ()

        val parents =
          TPTP_Reconstruct.parents_of_node fms node_name
          |> rev
        val _ =
          if 1 <= length parents andalso length parents <= 8 then ()
          else error ("Unexpected number of parents: " ^ Int.toString (length parents))

        val assump_parents =
          filter (is_assumption_node fms) parents
          |> map (Binding.name
              #> Binding.qualify false
                ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)
              #>
                (*NOTE need to do this, it seems, despite not needing
                       it at the time that the bindings were made.*)
                Binding.qualify false (Context.theory_name thy)
              #>
              (*FIXME this next phrase doesn't feel idiomatic*)
              (fn b =>
                  ((Binding.path_of #> map fst) b
                   @ [Binding.name_of b]))
              #> Long_Name.implode)
        val _ =
          if debug_mode (Context.Theory thy |> Context.proof_of) then
            warning ("assump_parents = " ^ @{make_string} assump_parents)
          else ()
        val _ =
          if 0 <= length assump_parents andalso length assump_parents <= 6 then ()
          else error ("Unexpected number of assump_parents: " ^ Int.toString (length assump_parents))

        val nassum_parents = filter (not o is_assumption_node fms) parents
        val _ =
          if debug_mode (Context.Theory thy |> Context.proof_of) then
            warning ("nassum_parents = " ^ @{make_string} nassum_parents)
          else ()
        val _ =
          if 0 <= length nassum_parents andalso length nassum_parents <= 2 then ()
          else error ("Unexpected number of nassum_parents: " ^ Int.toString (length nassum_parents))

        (*The shape of the inference rule we get depends on the number
          and "kind" of parents the node has -- by "kind" of parents recall
          the assumption vs inference node distinction mentioned above.*)
        val fmla' =
          case (length assump_parents, length nassum_parents) of
            (2, 0) =>
              mk_conflict
               (List.nth (assump_parents, 0))
               (List.nth (assump_parents, 1))
          | (1, 0) =>
              (*used by tab_refl and tab_false*)
              mk_exfalso
               (List.nth (assump_parents, 0))
          | (2, 1) =>
              (*used by tab_conflict and tab_delta,
                but owing to the difference shapes
                of rules involving 3 assumptions,
                we should look at the name of the interence rule*)
              (case TPTP_Reconstruct.inference_name_of_node fms node_name of
                NONE => error "Expected node to have a name"
              | SOME "tab_delta" =>
                  arbitrary_elim
                   (List.nth (assump_parents, 0))
                   (List.nth (assump_parents, 1))
              | SOME _ =>
                  (*used by tab_conflict and tab_negall*)
                  mk_elim_inference1
                   (*Don't pass known lemmas through to
                     the formula-forming function.*)
                   (let val p = List.nth (assump_parents, 0)
                    in if is_known_lemma (Long_Name.base_name p) then
                      NONE else SOME p end)
                   (List.nth (assump_parents, 1)))
          | (2, 2) =>
              (*only used by tab_cut*)
                mk_cut
                  (List.nth (assump_parents, 0))
                  (List.nth (assump_parents, 1))
          | (3, 1) =>
              (*used by tab_and, tab_trans, and
                tab_mat (when it doesn't branch),
                but owing to the difference shapes
                of rules involving 3 assumptions,
                we should look at the name of the interence rule*)
              (case TPTP_Reconstruct.inference_name_of_node fms node_name of
                NONE => error "Expected node to have a name"
              | SOME "tab_trans" =>
                  mk_elim_inference2_1
                   (List.nth (assump_parents, 0))
                   (List.nth (assump_parents, 1))
                   (List.nth (assump_parents, 2))
              | SOME "tab_mat" =>
                  mk_tabmat_3
                   (List.nth (assump_parents, 0))
                   (List.nth (assump_parents, 1))
                   (List.nth (assump_parents, 2))
              | SOME _ =>
                  mk_elim_inference1_2
                   (List.nth (assump_parents, 0))
                   (List.nth (assump_parents, 1))
                   (List.nth (assump_parents, 2)))
          | (3, 2) =>
              (*used by tab_imp and tab_mat
                but owing to the difference shapes
                of rules involving 3 assumptions,
                we should look at the name of the interence rule*)
              (case TPTP_Reconstruct.inference_name_of_node fms node_name of
                NONE => error "Expected node to have a name"
              | SOME _ =>
                 mk_elim_inference2
                  (List.nth (assump_parents, 0))
                  (List.nth (assump_parents, 1))
                  (List.nth (assump_parents, 2)))
          | (4, 2) =>
            (*used by tab_mat*)
            mk_elim_inference2_2
             (List.nth (assump_parents, 0))
             (List.nth (assump_parents, 1))
             (List.nth (assump_parents, 2))
             (List.nth (assump_parents, 3))
          | (5, 2) =>
              (*used by tab_be*)
              mk_elim_inference2_5
               (List.nth (assump_parents, 0))
               (List.nth (assump_parents, 1))
               (List.nth (assump_parents, 2))
               (List.nth (assump_parents, 3))
               (List.nth (assump_parents, 4))
          | (6, 2) =>
              (*used by tab_con*)
              mk_elim_inference2_6
               (List.nth (assump_parents, 0))
               (List.nth (assump_parents, 1))
               (List.nth (assump_parents, 2))
               (List.nth (assump_parents, 3))
               (List.nth (assump_parents, 4))
               (List.nth (assump_parents, 5))
          | (assumps, nassums) =>
              error ("Unexpected number of node " ^ node_name ^ " parents: " ^
                     string_of_int assumps ^ " assumptions, " ^
                     string_of_int nassums ^ " non-assumptions")

        (*Transform the node's formula -- this is the main
          operation of this transformation*)
        val node' =
          (node_name,
           {role = #role node,
            fmla = fmla',
            source_inf_opt = #source_inf_opt node})
      in
        traverse (rest @ nassum_parents) (node' :: acc)
    end

  (*Recall that the 'traverse' function won't work on
    assumptions; but we need to include the assumption nodes
    in the pannot produced by refine_conclusion otherwise
    things will break down the line.
    Here we gather them to make sure that they get included.*)
  val assumptions =
    filter (fst #> is_assumption_node fms) fms

in
  {problem_name = #problem_name pannot,
   skolem_defs = #skolem_defs pannot,
   defs = #defs pannot,
   axs = #axs pannot,
   meta = traverse [start_node_id] [] @ assumptions}
end
*}


section "Loading"

ML {*
(*Preprocessing carried out on a Satallax proof during
  its importation.*)
fun satallax_on_load (pannot : TPTP_Reconstruct.proof_annotation) thy =
  let
    val pannot' = (*remove_theorem_roled_AFs*) pannot
    val thy' =
      absorb_skolem_equations "tab_negall" negall_to_witness pannot' thy
      |> absorb_skolem_equations "tab_ex" ex_to_witness pannot'
    val assumption_defs = extract_assumptions pannot'

    val _ =
      if debug_mode (Context.Theory thy |> Context.proof_of) then
        fold (fn (n, _) => fn _ => warning ("Assumption:" ^ n)) assumption_defs ()
      else ()

    (*Add the assumption definitions to the theory*)
    val thy'' =
      fold (fn (name, info) => fn thy' =>
        absorb_assumption_def name (#fmla info) pannot thy')
       assumption_defs thy'
    val pannot'' =
      case TPTP_Reconstruct.filter_by_name starting_nodeid_of_proof (#meta pannot') of
        [(node_id, _)] =>
          (@{assert} (node_id = starting_nodeid_of_proof);
           refine_conclusion starting_nodeid_of_proof pannot' thy)
      | _ => error ""
  in
    (pannot'', thy'')
  end
*}


section "Emulation"

subsection "Tactic support"

ML {*
(*Given the name of an assumption used in a Satallax
  proof, obtain its definition (i.e., the meta-equality
  relating it to a formula) and return the tactic that
  expands occurrences of the constant (that abbreviates
  the assumption) into the formula -- i.e., unfolds
  its definition.*)
fun unfold_hyp_tac (pannot : TPTP_Reconstruct.proof_annotation) hyp_name ctxt : tactic =
  let
    val thy = Proof_Context.theory_of ctxt
  in
    (*First we need to form the binding, using which
      we will query the system to get the meta-equality
      that we'll use for the unfolding.*)
    Binding.name (Long_Name.base_name hyp_name)
    |> Binding.suffix_name def_suffix
    |> Binding.qualify false
      ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)
     (*NOTE need to do this, despite not needing it at the time
            that the bindings were made.*)
    |> Binding.qualify false (Context.theory_name thy)
    (*FIXME this next phrase doesn't feel idiomatic*)
    |> (fn b =>
          ((Binding.path_of #> map fst) b
           @ [Binding.name_of b]))
    |> Long_Name.implode

    |> (fn s =>
          (if debug_mode (Context.Theory thy |> Context.proof_of) then
            warning ("unfold_hyp_tac <- " ^ s)
          else ();
          s))

    (*Querying the system to get the meta-equality, and 
      use it for unfolding.*)
    |> Thm.axiom thy
    |> single
    |> unfold_tac ctxt
  end

(*Unfold a list of hypotheses in order*)
fun unfold_hyplist_tac pannot hyp_names ctxt : tactic =
  map (fn hyp_name => unfold_hyp_tac pannot hyp_name ctxt) hyp_names
  |> EVERY
*}


subsection "Emulation tactics"

ML {*
fun TabDec_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TabDec} THEN'
    Method.assm_tac ctxt)
*}

ML {*
fun TabFE_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TFE} THEN'
    Method.assm_tac ctxt)
*}

ML {*
fun TabFQ_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TFQ} THEN'
    Method.assm_tac ctxt)
*}

ML {*
fun TabDelta_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TabDelta} THEN'
    Method.assm_tac ctxt)
*}

ML {*
fun TabDN_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TDN} THEN'
    Method.assm_tac ctxt)
*}

ML {*
fun TabConflict_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TabConflict}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabRefl_tac pannot Ha ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (eresolve_tac ctxt @{thms TabRefl}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabFalse_tac pannot Ha ctxt : tactic =
(*Again, it seems that the unfolding tactic is stronger
  than i thought -- it doesn't simply do unfolding and
  stop there, so i added the TRY steps*)
  unfold_hyp_tac pannot Ha ctxt
  THEN ((TRY o HEADGOAL) (eresolve_tac ctxt @{thms TabFalse}))
  THEN ((TRY o REPEAT o HEADGOAL) (Method.assm_tac ctxt))
*}

ML {*
fun TabImp_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TImp}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabNAnd_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TNAnd}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabOr_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TOr}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun eq_ind_r_tac pannot Ha ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm eq_ind_r}) THEN
  HEADGOAL (Method.assm_tac ctxt)

fun eq_ind_tac pannot Ha ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm eq_ind}) THEN
  HEADGOAL (Method.assm_tac ctxt)

fun eq_sym_tac pannot Ha ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm eq_sym}) THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabAnd_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc] ctxt THEN
  HEADGOAL (etac @{thm TAnd}) THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabTrans_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc] ctxt THEN
  HEADGOAL (etac @{thm Ttrans}) THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabNor_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc] ctxt THEN
  HEADGOAL (etac @{thm TNor}) THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
(*blast_tac is used here to deal with syntactic changes
  made by Satallax (e.g., "~ false" to "true") -- though
  seemingly trivial, this is syntactically different than
  what's expected. Rather than deal with this using complex
  machinery as with LEO-II's extcnf, we rely on blast_tac
  to hopefully emulate this inference.*)
fun TabNegImp_tac pannot Ha Hb Hc ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc] ctxt THEN
  ((HEADGOAL (etac @{thm TNegImp}) THEN
    HEADGOAL (Method.assm_tac ctxt)) ORELSE HEADGOAL (blast_tac ctxt))
*}

ML {*
fun TabMat_tac pannot Ha Hb Hc Hd ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc, Hd] ctxt THEN
  HEADGOAL (etac @{thm TMat}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabCon_tac pannot Ha Hb Hc Hd He Hf ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc, Hd, He, Hf] ctxt THEN
  HEADGOAL (etac @{thm TCON}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabCut_tac pannot Ha Hb ctxt : tactic =
(*This was the original tactic, but didn't work; it seems we really need to help it out by instantiating:
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (etac @{thm TCut}) THEN
  HEADGOAL (Method.assm_tac ctxt)
*)
  HEADGOAL (Rule_Insts.eres_inst_tac ctxt [((("Q", 0), Position.none), Ha)] [] @{thm TCut})
  THEN HEADGOAL (Method.assm_tac ctxt)
  THEN unfold_hyplist_tac pannot [Ha, Hb] ctxt
  THEN HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabNegAll_tac pannot Ha Hb skolem_const_name skolem_eq_thm_name ctxt i : tactic =
  let
    val thy = Proof_Context.theory_of ctxt
  in
    unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
    Rule_Insts.eres_inst_tac ctxt [((("sk", 0), Position.none), skolem_const_name)] []
              @{thm TNegAll} i THEN
    rtac (Thm.axiom thy skolem_eq_thm_name) i THEN
    Method.assm_tac ctxt i
  end
*}

ML {*
fun TabEx_tac pannot Ha Hb skolem_const_name skolem_eq_thm_name ctxt i : tactic =
  let
    val thy = Proof_Context.theory_of ctxt
  in
    unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
    Rule_Insts.eres_inst_tac ctxt [((("sk", 0), Position.none), skolem_const_name)] []
              @{thm TEx} i THEN
    rtac (Thm.axiom thy skolem_eq_thm_name) i THEN
    Method.assm_tac ctxt i
  end
*}

ML {*
fun TabNegAll_notfree_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (etac @{thm TNegAll_notfree}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabEx_notfree_tac pannot Ha Hb ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (etac @{thm TEx_notfree}) THEN
  (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
*}

ML {*
fun TabAll_tac pannot Ha Hb term_string ctxt i : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  Rule_Insts.eres_inst_tac ctxt [((("y", 0), Position.none), term_string)] []
            @{thm TabAll} i THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  Method.assm_tac ctxt i
*}

ML {*
fun TabNegEx_tac pannot Ha Hb term_string ctxt i : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  Rule_Insts.eres_inst_tac ctxt [((("y", 0), Position.none), term_string)] []
            @{thm TNegEx} i THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  Method.assm_tac ctxt i
*}

ML {*
fun TabBE_tac pannot Ha Hb Hc Hd He ctxt : tactic =
(*There was some sequencing anomaly when using this tactic --
  the goal was solved prematurely!

  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TBE}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hd ctxt THEN
  unfold_hyp_tac pannot He ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*)
  unfold_hyplist_tac pannot [Ha, Hb, Hc, Hd, He] ctxt THEN
  HEADGOAL (etac @{thm TBE}) THEN
  REPEAT (HEADGOAL (Method.assm_tac ctxt))
*}

ML {*
fun TabBQ_tac pannot Ha Hb Hc Hd He ctxt : tactic =
  unfold_hyplist_tac pannot [Ha, Hb, Hc, Hd, He] ctxt THEN
  HEADGOAL (etac @{thm TBQ}) THEN
  REPEAT (HEADGOAL (Method.assm_tac ctxt))
*}

ML {*
fun TabIff_tac pannot Ha Hb Hc Hd He ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TIff}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hd ctxt THEN
  unfold_hyp_tac pannot He ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*}

ML {*
fun TabNIff_tac pannot Ha Hb Hc Hd He ctxt : tactic =
  unfold_hyp_tac pannot Ha ctxt THEN
  HEADGOAL (etac @{thm TNIff}) THEN
  unfold_hyp_tac pannot Hb ctxt THEN
  unfold_hyp_tac pannot Hc ctxt THEN
  HEADGOAL (Method.assm_tac ctxt) THEN
  unfold_hyp_tac pannot Hd ctxt THEN
  unfold_hyp_tac pannot He ctxt THEN
  HEADGOAL (Method.assm_tac ctxt)
*}


subsection "TMat3"

ML {*
(*FIXME move to library*)
(*Iteratively applies Logic.dest_implies to consequents.*)
fun dest_implies_lst acc t =
  let
    val (t1, t2) = Logic.dest_implies t
  in
    dest_implies_lst (t1 :: acc) t2
  end
  (*NOTE we don't add "t" to "acc" since "t"
         must be the conclusion, not a premise.*)
  handle TERM ("dest_implies", _) => rev acc
*}

ML {*
(*FIXME move to library*)
(* Extract a specified goal from the given proof state.
   subgoal_no is 0-indexed*)
fun extract_subgoal_from_state (subgoal_no : int) (st : thm) : term option =
let
  val subgoals =
    Thm.prop_of st
    |> Logic.strip_horn
    |> fst
in
  if length subgoals <= subgoal_no then NONE
  else
    rpair subgoal_no subgoals
    |> uncurry nth
    |> strip_top_all_vars []
    |> snd
    |> SOME
end
*}

ML {*
(*FIXME move to library*)
(* Extract a specified premise from a specified goal from the given proof state.
   Both subgoal_no and premise_no are 0-indexed*)
fun extract_premise_from_subgoal_from_state (premise_no : int) (subgoal_no : int) (st : thm) : term option =
  case extract_subgoal_from_state subgoal_no st of
    NONE => NONE
  | SOME t =>
    let
      val premises = dest_implies_lst [] t
    in
      if length premises <= premise_no then NONE
      else
        nth premises premise_no
        |> SOME
    end

(*FIXME move to library*)
local
  fun get_g_f_comb_from_propfmla g f (t : term) : term =
    HOLogic.dest_Trueprop t
    |> g
    |> dest_comb
    |> f
in
  (*Returns the result of applying a function f to
    a combination that's wrapped in a Trueprop.
    "f" is usually fst or snd*)
  val get_f_comb_from_propfmla =
    get_g_f_comb_from_propfmla (fn x => x)
  (*Like get_f_comb_from_propfmla, but targets a
    negated formula*)
  val get_f_comb_from_negated_propfmla =
    get_g_f_comb_from_propfmla HOLogic.dest_not
end
*}

ML{*
(*Instantiates TMat3 before applying it. The instantiation
  is meant to diminish the scope of matching.*)
fun customise_TMat3_tac ctxt st  =
let
  val prem1 =
    extract_premise_from_subgoal_from_state 0 0 st
  val prem2 =
    extract_premise_from_subgoal_from_state 1 0 st
  val p1 =
    case prem1 of
      NONE => NONE (*error "Could not extract headconst from first premise"*)
    | SOME t =>
      get_f_comb_from_propfmla fst t
      |> (dest_Const #> fst)
      |> SOME
  val p2 =
    case prem2 of
      NONE => NONE
    | SOME t =>
      ((get_f_comb_from_negated_propfmla fst t
        |> (dest_Const #> fst)
        |> SOME)
        (*In case we got the order of the premises
          wrong, and the negated formula isn't in the
          premise_no we expect it to be, then just fail
          the tactic so that a different approach
          (probably consisting of a permutation of
          premises) can be attempted.*)
        handle (TERM ("dest_not", _)) => NONE)
in
  if p2 = NONE orelse p1 = NONE (*we shouldn't really expect p1=NONE, but let's fail gracefully*) orelse
    (is_some p1 andalso is_some p2 andalso p2 <> p1) then no_tac st
  else
    (@{assert} (is_some p1 andalso is_some p2 andalso p2 = p1);
     let
       val p = the p1
       val s = get_f_comb_from_propfmla snd (the prem1)
         |> term_to_string ctxt
       val t = get_f_comb_from_negated_propfmla snd (the prem2)
         |> term_to_string ctxt
       val instantiation =
         Rule_Insts.eres_inst_tac ctxt
           [((("p", 0), Position.none), p),
            ((("s", 0), Position.none), s),
            ((("t", 0), Position.none), t)]
           [] @{thm TMat3}
       val tac =
         HEADGOAL instantiation
         THEN (REPEAT o HEADGOAL) (Method.assm_tac ctxt)
     in
       tac st
     end)
end
*}

ML {*
fun TabMat3_tac pannot Ha Hb Hc ctxt =
  unfold_hyplist_tac pannot [Ha, Hb, Hc] ctxt THEN
  (* Extract the first two premises (unfoldings of Ha and Hb)
     extract the rator & rand of each
     use these to instantiate TMat3 accurately*)
  customise_TMat3_tac ctxt
*}


subsection "Rewriting"

ML {*
(*FIXME move to TPTP_Proof*)
fun extract_useful_info (source_info_opt : TPTP_Proof.source_info option) : TPTP_Proof.useful_info_as list =
  case source_info_opt of
    NONE => []
  | SOME (TPTP_Proof.File _) => []
  | SOME (TPTP_Proof.Inference (_, useful_info, _)) => useful_info
*}

ML {*
(*Given the "useful info" annotation for a rewriting-based inference,
  extract the function that will be applied to both sides of a
  rewrite equation. Such equations include pre_eq_neg_neg_id --
  see Satallax.thy*)
val extract_fn_for_rewrite : TPTP_Proof.useful_info_as list -> TPTP_Syntax.tptp_formula =
  map (fn useful_info =>
    case useful_info of
      TPTP_Proof.Other (
        TPTP_Syntax.General_Data
         (TPTP_Syntax.Application
           (_(*NOTE could check that the inference rule is what we expect it to be*),
            [TPTP_Syntax.General_Data (TPTP_Syntax.Atomic_Word "leibnizp"),
             TPTP_Syntax.General_Data
              (TPTP_Syntax.Formula_Data
                (TPTP_Syntax.THF, fmla))]))) => SOME fmla
    | _ => NONE)
  #> fold_options
  #> (fn result =>
    case result of
      [fmla] => fmla
    | _ => error "There should only be a single rewriting function")
*}

ML {*
(*Extract the function that will be applied during a rewriting step
  -- this will be applied to both sides of equations like pre_eq_neg_neg_id
  (depending on the rewriting sought) in Satallax.thy to produce theorems
  that will be applied to tab_rew, thus giving us*)
(*NOTE I don't explicitly check that node_id is a rewriting inference
       -- doing this check would be an improvement. The check is done
       implicitly when trying to extract the function from the
       "useful info" annotation.
*)
fun extract_rewrite_fn ctxt (pannot : TPTP_Reconstruct.proof_annotation) (node_id : string) : term * theory =
let
(*FIXME DRY -- parts of this block were copied from emulate_id*)
  val prob_name = #problem_name pannot
  (*FIXME get config from the envir or make it parameter*)
  val config =
    {cautious = true,
     problem_name = SOME prob_name}
  val thy = Proof_Context.theory_of ctxt
  val (type_map, const_map) =
      case AList.lookup (op =) (TPTP_Interpret.get_manifests thy) prob_name of
          NONE => raise (TPTP_Reconstruct.MANIFEST (prob_name, "Problem details not found in interpretation manifest"))
        | SOME (type_map, const_map, _) => (type_map, const_map)
  val fmla =
    TPTP_Reconstruct.node_info (#meta pannot) #source_inf_opt node_id
    |> extract_useful_info
    |> extract_fn_for_rewrite
in
  TPTP_Interpret.interpret_formula
   config TPTP_Syntax.THF
   const_map [] type_map fmla thy
end
*}

ML {*
(*Given an equation, and a function term to be applied to both
  sides of the equation, it computes an instance of tab_rew
  (from Satallax.thy).*)
fun tab_rew_instance ctxt (eq_thm : thm) (f_t : term) : thm =
let
  val f_s = term_to_string ctxt f_t
  (*justifies rewriting the whole formula -- while f_t only
    justifies the rewriting of the subformula within.*)
  val rewrite_thm =
    (@{thm arg_cong} OF [eq_thm])
    |> Rule_Insts.where_rule ctxt
        [((("f", 0), Position.none), f_s)] []
in
  (*Apply rewrite_thm to tab_rew, to obtain an instance
    of the rewrite rule we seek (specialised as needed).*)
  @{thm tab_rew} OF [rewrite_thm]
end
*}

ML {*
(*Given an equation, and a function term to be applied to both
  sides of the equation, it computers an instance of tab_rew
  (from Satallax.thy) and uses it to build a tactic.*)
fun tab_rewrite_tac Ha Hb ctxt (eq_thm : thm)
  (pannot : TPTP_Reconstruct.proof_annotation) (f_t : term) : tactic =
let
  val tab_rew_instance = tab_rew_instance ctxt eq_thm f_t
in
  (*Given tab_rew_instance, build a tactic that proves it.
    This is a bit redundant, given that we have tab_rew_instance
    already, and the tactic proves tab_rew_instance using itself,
    but I do this to have a uniform interface for emulation tactics
    -- so they will all go through the "emulate_id" function.*)
(*
  HEADGOAL (etac tab_rew_instance)
  THEN (REPEAT (HEADGOAL (Method.assm_tac ctxt)))
*)
  unfold_hyplist_tac pannot [Ha, Hb] ctxt THEN
  HEADGOAL (etac tab_rew_instance) THEN
  HEADGOAL (Method.assm_tac ctxt)
end
*}


subsection "Emulation abstraction functions"

ML {*
(*As a fall-back, we unfold all definitions and apply some internal automation*)
fun fallback pannot ctxt parents =
  unfold_hyplist_tac pannot parents ctxt THEN
  ((*(REPEAT o ALLGOALS) (blast_tac ctxt)
   APPEND*) ALLGOALS (Metis_Tactic.metis_tac [] ATP_Problem_Generate.combs_or_liftingN ctxt []))
*}

ML {*
(*Given the name of an inference rule used by Satallax, and parameters
  to that inference, this function checks if it has the right parameters
  and then calls the emulation function for that particular inference.*)
fun emulate (inference_name : string) (pannot : TPTP_Reconstruct.proof_annotation)
  (hyp_names : string list)
  (term_args : string list) ctxt : tactic =
  case inference_name of
    "tab_conflict" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabConflict_tac pannot Ha Hb ctxt
      end
  | "tab_refl" =>
      let
        val Ha = case hyp_names of
            [Ha] => Ha
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabRefl_tac pannot Ha ctxt
      end
  | "tab_false" =>
      let
        val Ha = case hyp_names of
            [Ha] => Ha
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabFalse_tac pannot Ha ctxt
      end
  | "tab_dec" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabDec_tac pannot Ha Hb ctxt
      end
  | "tab_delta" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabDelta_tac pannot Ha Hb ctxt
      end
  | "tab_fe" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabFE_tac pannot Ha Hb ctxt
      end
  | "tab_fq" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabFQ_tac pannot Ha Hb ctxt
      end
  | "tab_dn" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabDN_tac pannot Ha Hb ctxt
      end
  | "tab_cut" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabCut_tac pannot Ha Hb ctxt
      end
  | "tab_imp" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabImp_tac pannot Ha Hb Hc ctxt
      end
  | "tab_nand" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabNAnd_tac pannot Ha Hb Hc ctxt
      end
  | "tab_or" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabOr_tac pannot Ha Hb Hc ctxt
      end
  | "tab_be" =>
      let
        val (Ha, Hb, Hc, Hd, He) = case hyp_names of
            [Ha, Hb, Hc, Hd, He] => (Ha, Hb, Hc, Hd, He)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabBE_tac pannot Ha Hb Hc Hd He ctxt
      end
  | "tab_bq" =>
      let
        val (Ha, Hb, Hc, Hd, He) = case hyp_names of
            [Ha, Hb, Hc, Hd, He] => (Ha, Hb, Hc, Hd, He)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabBQ_tac pannot Ha Hb Hc Hd He ctxt
      end
  | "tab_iff" =>
      let
        val (Ha, Hb, Hc, Hd, He) = case hyp_names of
            [Ha, Hb, Hc, Hd, He] => (Ha, Hb, Hc, Hd, He)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabIff_tac pannot Ha Hb Hc Hd He ctxt
      end
  | "tab_negiff" =>
      let
        val (Ha, Hb, Hc, Hd, He) = case hyp_names of
            [Ha, Hb, Hc, Hd, He] => (Ha, Hb, Hc, Hd, He)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabNIff_tac pannot Ha Hb Hc Hd He ctxt
      end
  | "tab_known" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in (case Long_Name.base_name Ha of
              "eq_ind_r" => eq_ind_r_tac pannot Hb ctxt
            | "eq_ind" => eq_ind_tac pannot Hb ctxt
            | "eq_sym" => eq_sym_tac pannot Hb ctxt
            | _ => error ("Unrecognised lemma: " ^ Ha))
      end
  | "tab_trans" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabTrans_tac pannot Ha Hb Hc ctxt
      end
  | "tab_and" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabAnd_tac pannot Ha Hb Hc ctxt
      end
  | "tab_nor" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabNor_tac pannot Ha Hb Hc ctxt
      end
  | "tab_negimp" =>
      let
        val (Ha, Hb, Hc) = case hyp_names of
            [Ha, Hb, Hc] => (Ha, Hb, Hc)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabNegImp_tac pannot Ha Hb Hc ctxt
      end
  | "tab_mat" =>
      let
        (*tab_mat sometimes has 3 associated hypothesis parents
          (and a single non-assumption parent), or
          4 hypothesis parents (and 2 non-assumption parents).
          The first is a slight optimisation of the latter,
          where the discrimination is trivial (i.e., p \<noteq> p).
          The reconstruction is flexible to whichever of these
          2 variants of tab_mat is used by Satallax.*)
        val (Ha, Hb, Hc, Hd_opt) = case hyp_names of
            [Ha, Hb, Hc, Hd] => (Ha, Hb, Hc, SOME Hd)
          | [Ha, Hb, Hc] => (Ha, Hb, Hc, NONE)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        case Hd_opt of
          NONE =>
            (*NOTE I permute Ha and Hb since I'm not sure which
                   one is needed in this instance.*)
            TabMat3_tac pannot Ha Hb Hc ctxt
            ORELSE TabMat3_tac pannot Hb Ha Hc ctxt
(*            ORELSE fallback pannot ctxt hyp_names*)
        | SOME Hd => TabMat_tac pannot Ha Hb Hc Hd ctxt
      end
  | "tab_con" =>
      let
        val (Ha, Hb, Hc, Hd, He, Hf) = case hyp_names of
            [Ha, Hb, Hc, Hd, He, Hf] => (Ha, Hb, Hc, Hd, He, Hf)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabCon_tac pannot Ha Hb Hc Hd He Hf ctxt
      end
  | "tab_negall" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val (skolem_const_name, skolem_eq_thm_name) = case term_args of
            [skolem_const_name, skolem_eq_thm_name] => (skolem_const_name, skolem_eq_thm_name)
          | _ => error "Incorrect number of term arguments."
      in
        HEADGOAL (TabNegAll_tac pannot Ha Hb skolem_const_name skolem_eq_thm_name ctxt)
      end
  | "tab_ex" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val (skolem_const_name, skolem_eq_thm_name) = case term_args of
            [skolem_const_name, skolem_eq_thm_name] => (skolem_const_name, skolem_eq_thm_name)
          | _ => error "Incorrect number of term arguments."
      in
        HEADGOAL (TabEx_tac pannot Ha Hb skolem_const_name skolem_eq_thm_name ctxt)
      end
  
  | "tab_negall_notfree" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabNegAll_notfree_tac pannot Ha Hb ctxt
      end
  | "tab_ex_notfree" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val _ = case term_args of
            [] => ()
          | _ => error "Incorrect number of term arguments."
      in
        TabEx_notfree_tac pannot Ha Hb ctxt
      end
  
  | "tab_all" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val instance = case term_args of
            [instance] => instance
          | _ => error "Incorrect number of term arguments."
      in
        HEADGOAL (TabAll_tac pannot Ha Hb instance ctxt)
      end
  | "tab_negex" =>
      let
        val (Ha, Hb) = case hyp_names of
            [Ha, Hb] => (Ha, Hb)
          | _ => error "Incorrect number of hypotheses."
        val instance = case term_args of
            [instance] => instance
          | _ => error "Incorrect number of term arguments."
      in
        HEADGOAL (TabNegEx_tac pannot Ha Hb instance ctxt)
      end
  | _ => error ("Unrecognised Satallax inference name: " ^ inference_name)
*}

ML {*
(*Emulate a particular node in the proof graph, identified by its id.
  Involves first extracting the hypotheses involved, and
  any term parameters, then calling the "emulate" function.*)
(*NOTE this function is similar to interpret_leo2_inference*)
fun emulate_id (node_id : string) (pannot : TPTP_Reconstruct.proof_annotation) ctxt : tactic =
let
  val fms = #meta pannot
  val inference_name =
    case TPTP_Reconstruct.inference_name_of_node fms node_id of
      NONE => error "Was expected to find name of a Satallax inference rule."
    | SOME name => name

  val _ =
    if debug_mode ctxt then
      warning (">>>" ^ node_id ^ " : " ^ inference_name)
    else ()

  val parents =
    (*Filter away non-assumptions*)
    TPTP_Reconstruct.parents_of_node fms node_id
    |> filter (is_assumption_node fms)
    |> rev

  val prob_name = #problem_name pannot
  (*FIXME get config from the envir or make it parameter*)
  val config =
    {cautious = true,
     problem_name = SOME prob_name}
  val thy = Proof_Context.theory_of ctxt
  val (type_map, const_map) =
      case AList.lookup (op =) (TPTP_Interpret.get_manifests thy) prob_name of
          NONE => raise (TPTP_Reconstruct.MANIFEST (prob_name, "Problem details not found in interpretation manifest"))
        | SOME (type_map, const_map, _) => (type_map, const_map)
  (*Extract binding info from parent details*)
  fun extract_bindings source_inf_opt =
    case source_inf_opt of
        NONE => []
      | SOME (TPTP_Proof.File _) => []
      | SOME (TPTP_Proof.Inference (_, _, parent_infos)) =>
        filter (fn parent_info =>
            case parent_info of
              TPTP_Proof.Parent _ => false
            | TPTP_Proof.ParentWithDetails (_, _) => true) parent_infos
        |> map (fn TPTP_Proof.ParentWithDetails (n, bindings) =>
            (n, map 
                (fn TPTP_Proof.Bind (var, fmla) =>
                  (var,
                   TPTP_Interpret.interpret_formula
                    config TPTP_Syntax.THF
                    const_map [] type_map fmla thy)) bindings)
                 | _ => error "Unexpected shape of parent detail")

  val _ =
    if debug_mode ctxt then
      warning ("parents = " ^ @{make_string} parents)
    else ()

  fun rewrite_using eq_thm =
    let
      val (Ha, Hb) =
        case parents of
          [Ha, Hb] => (Ha, Hb)
        | _ => error "Unexpected number of parents"
    in
      extract_rewrite_fn ctxt pannot node_id
      |> fst
      |> tab_rewrite_tac Ha Hb ctxt eq_thm pannot
    end

  val tac =
    if List.exists (fn inf_name => inf_name = inference_name)
    ["tab_dec", "tab_fe", "tab_dn", "tab_conflict", "tab_imp",
     "tab_nand", "tab_or", "tab_be", "tab_known", "tab_and",
     "tab_nor", "tab_negimp", "tab_bq", "tab_iff", "tab_negiff",
     "tab_fq", "tab_trans", "tab_mat", "tab_con", "tab_cut",
     "tab_refl", "tab_false", "tab_delta"] then
       emulate inference_name pannot parents [] ctxt
    else if List.exists (fn inf_name => inf_name = inference_name)
    ["tab_negall", "tab_ex"] then
       let
         (*We extract the info we needed from known places in the theory.*)
         val thm_name =
           Binding.name (Long_Name.base_name node_id)
           |> Binding.suffix_name skolemeq_suffix
           |> Binding.qualify false
             ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)
          
            (*NOTE need to do this, despite not needing it at the time
                   that the bindings were made.*)
           |> Binding.qualify false (Context.theory_name thy)
            (*FIXME this next phrase doesn't feel idiomatic*)
           |> (fn b =>
                  ((Binding.path_of #> map fst) b
                   @ [Binding.name_of b])
           |> Long_Name.implode)
       
         val const_name =
           let
             val concl =
               Thm.axiom thy thm_name
               |> Thm.concl_of
           in
             if concl = demmy_eq then NONE
             else
               HOLogic.dest_Trueprop concl
               |> HOLogic.dest_eq
               |> fst
               |> dest_Const
               |> fst
               |> SOME
           end
       in
        case const_name of
          NONE =>
            emulate (inference_name ^ "_notfree") pannot parents [] ctxt
        | SOME const_name =>
            emulate inference_name pannot parents [const_name, thm_name] ctxt
       end
    
    else if List.exists (fn inf_name => inf_name = inference_name)
    ["tab_all", "tab_negex"] then
       let
         val sio = TPTP_Reconstruct.node_info fms #source_inf_opt node_id
         val term_inst =
           case extract_bindings sio of
             (*I think all other data, including the returned
               theory, can be discarded, since it shouldn't be
               any different from the current theory.*)
             [(_, [(_, (fmla, _))])] => fmla
           | _ => error "Unexpected binding shape"
       in
          emulate inference_name pannot parents [term_to_string ctxt term_inst] ctxt
       end
    else
     case inference_name of
       "tab_rew_dn" => rewrite_using @{thm pre_eq_neg_neg_id}
     | "tab_rew_and" =>
         rewrite_using @{thm pre_eq_and_imp}
         ORELSE rewrite_using @{thm pre_eq_and_nor}
     | "tab_rew_leib1" => rewrite_using @{thm pre_eq_leib1}
     | "tab_rew_leib2" => rewrite_using @{thm pre_eq_leib2}
     | "tab_rew_leib3" => rewrite_using @{thm pre_eq_leib3}
     | "tab_rew_leib4" => rewrite_using @{thm pre_eq_leib4}
     | "tab_rew_sym" => rewrite_using @{thm pre_eq_sym_eq}
     | "tab_rew_iff" => rewrite_using @{thm pre_eq_iff}
     | "tab_rew_true" => rewrite_using @{thm pre_eq_true}
     | "tab_rew_ex" => rewrite_using @{thm pre_eq_exists_nforall}
     | "tab_rew_eta" => rewrite_using @{thm pre_eq_eta}
     | "tab_rew_or" =>
         rewrite_using @{thm pre_eq_or_nand}
         ORELSE rewrite_using @{thm pre_eq_or_imp}
     | _ => error ("Unrecognised Satallax inference name: " ^ inference_name)
in
(*  fallback pannot ctxt (filter (not o is_known_lemma) parents)
  ORELSE*) tac
end
*}

(*Emulation wrapper: given an inference node it generates
  a meta-theorem that validates that inference.*)
ML {*
fun validate_node node_id (pannot : TPTP_Reconstruct.proof_annotation) ctxt =
  let
    val t = TPTP_Reconstruct.node_info (#meta pannot) #fmla node_id

    val _ =
      if debug_mode ctxt then
        warning ("Goal for " ^ node_id ^ ":" ^ term_to_string ctxt t)
      else ()

    val thm = Goal.prove ctxt [] [] t
      (fn _ =>
        let
          val fms = #meta pannot
          val parents =
            TPTP_Reconstruct.parents_of_node fms node_id
            |> filter (is_assumption_node fms)
            |> rev
          val fallback_tac =
            case TPTP_Reconstruct.inference_name_of_node fms node_id of
              NONE => error "Was expected to find name of a Satallax inference rule."
            (*We first try to emulate tab_mat and tab_dec using Metis,
              since tab_mat is more difficult to emulate directly*)
            | SOME "tab_mat" =>
              fallback pannot ctxt (filter (not o is_known_lemma) parents)
            | SOME "tab_dec" =>
              fallback pannot ctxt (filter (not o is_known_lemma) parents)
            | SOME "tab_conflict" =>
              fallback pannot ctxt (filter (not o is_known_lemma) parents)
            | SOME _ => no_tac
        in
          fallback_tac ORELSE
          emulate_id node_id pannot ctxt
        end)
  in
    thm
  end
*}


section "Plan generation and execution"

ML {*
(*Traverses the proof graph, generating a cut program.*)
fun make_plan (node_id : string) (pannot : TPTP_Reconstruct.proof_annotation) : string list =
  let
    val fms = #meta pannot
    val parents =
      (*Filter away assumptions*)
      TPTP_Reconstruct.parents_of_node fms node_id
      |> filter (not o is_assumption_node fms)
      |> rev
    val rest =
      if parents = [] then
        [terminal_node]
      else
        fold (fn p => fn acc =>
          acc @ make_plan p pannot) parents []
  in
    node_id :: rest
  end
*}

ML {*
(*Map a cut program into a tactic. A plan contains
  references to nodes in the proof graph, in
  in the order in which they need to be applied
  (once emulated) in order to reconstruct the proof*)
fun cut_machine pannot ctxt plan =
let
  (*Generate a tactic that tries to solve k
    subgoals by assumption. This is done for cleaning
    up after rules are applied, when they leave
    dangling assumption steps.*)
  fun k_assumption_applications_tac k =
    (*if k = 0 then this should behave the
      same as EVERY [], which is same as all_tac*)
    1 upto k
    |> map (fn _ => HEADGOAL (Method.assm_tac ctxt))
    |> EVERY

  (*generate the tactic for node_id, including cleanup*)
  fun step_tac node_id st =
    let
      val thm = validate_node node_id pannot ctxt
      (*the number of premises to clean up*)
      val cleanups =
        Thm.prop_of thm
        |> strip_top_all_vars []
        |> snd
        |> dest_implies_lst []
          (*Don't count ==> premises since they don't
            need to be cleaned up -- they are branches
            that need to be extended further.*)
        |> filter (fn t =>
            case t of
              (Const (@{const_name "Pure.imp"}, _) $ _ $ _) => false
            | _ => true)
        |> length

      val tac =
        HEADGOAL (etac thm)
        THEN (*TRY*) (k_assumption_applications_tac (cleanups - 1))
    in
      tac st
    end

  fun debug_tac node_id st =
    if debug_mode ctxt then
      (warning ("Doing " ^ node_id);
       print_tac ctxt ("Doing " ^ node_id) st)
    else all_tac st

in
  plan
  |> map (fn node_id =>
    if node_id = terminal_node then
      (TRY o HEADGOAL) k_assumption_applications_tac
    else
      debug_tac node_id THEN step_tac node_id (*ORELSE HEADGOAL (blast_tac ctxt) -- or use fallback*))
  |> EVERY
end
*}


section "Main"

ML {*
(*Given a pannot, form the initial proof goal*)
fun formulate_goal (pannot : TPTP_Reconstruct.proof_annotation) ctxt =
  let
    val fms = #meta pannot
    val (_, node) =
      TPTP_Reconstruct.filter_by_role TPTP_Syntax.Role_Negated_Conjecture fms
      |> the_single (*There should only be one such node*)

    (*We infer what the original conjecture was by looking
      add the negated conjecture, and removing the negation.*)
    val t =
      #fmla node
      |> HOLogic.dest_Trueprop
      |> HOLogic.dest_not
      |> HOLogic.mk_Trueprop
  in
    Goal.prove ctxt [] [] t
  end
*}

ML {*
(*FIXME this code's repeated several times, but often
  we apply Long_Name.base_name to "name"*)
fun form_binding (suffix_opt : string option) (thy_opt : theory option)
    (pannot : TPTP_Reconstruct.proof_annotation)
    (name : string) : Binding.binding =
  Binding.name name
  |> fold_opt (fn suffix => Binding.suffix_name suffix) suffix_opt
  |> Binding.qualify false
    ((TPTP_Problem_Name.mangle_problem_name o #problem_name) pannot)
   (*NOTE need to do this, despite not needing it at the time
          that the bindings were made.*)
  |> fold_opt (fn thy => Binding.qualify false (Context.theory_name thy)) thy_opt
fun binding_to_string b =
  (*FIXME this function doesn't feel idiomatic*)
  ((Binding.path_of #> map fst) b @ [Binding.name_of b])        
  |> Long_Name.implode
*}

ML {*
(*Generates a tactic that, given the name of an
  assumption (which also doubles as the name of the
  constant that abbreviates it -- see absorb_assumption_def
  -- it will fold the abbreviated formula into
  an occurrence of the constant. This is only used
  once: at the very start of the proof, to fold the
  negated conjecture into the constant that abbreviates
  it.
*)
fun fold_assumption assump_name pannot ctxt st =
  let
    val thy = Proof_Context.theory_of ctxt
    val rule =
      form_binding (SOME def_suffix) (SOME thy) pannot assump_name
      |> binding_to_string
      |> Thm.axiom thy
      |> Thm.symmetric
      |> Conv.rewr_conv
      |> Conv.try_conv
    val conv =
      Conv.bottom_conv (K rule) ctxt
  in Conv.fconv_rule conv st
    |> Seq.single
  end

(*We initialise the proof by first using the
  contradiction rule, then folding the negated
  conjecture formula into the Boolean constant
  that abbreviates it.*)
fun initialise_proof_tac assump_name pannot ctxt =
  HEADGOAL (rtac @{thm ccontr})
  THEN fold_assumption assump_name pannot ctxt
*}

ML {*
(*Imports and reconstructs a Satallax proof.*)
(*NOTE would be nice to generalise
  TPTP_Reconstruct.reconstruct
  and TPTP_Reconstruct.make_skeleton, and
  factor common functionality in
  reconstruct_leo2 and reconstruct_satallax.
*)
fun reconstruct_satallax path thy =
  let
    val prob_file = Path.base path
    val dir = Path.dir path
    val thy' =
      TPTP_Reconstruct.import_thm true
       [dir, prob_file] path satallax_on_load thy
    val ctxt =
      Context.Theory thy'
      |> Context.proof_of
    val prob_name =
      Path.implode prob_file
      |> TPTP_Problem_Name.parse_problem_name
    val pannot =
      TPTP_Reconstruct.get_pannot_of_prob thy' prob_name

    val theorem =
      make_plan starting_nodeid_of_proof pannot
      |> (fn x =>
           (if debug_mode (Context.Theory thy |> Context.proof_of) then
              warning ("plan = " ^ @{make_string} x)
            else (); x))
      |> cut_machine pannot ctxt
      |> (fn x =>
           (if debug_mode (Context.Theory thy |> Context.proof_of) then
              warning ("made cut machine")
            else (); x))
      |> (fn tac =>
        formulate_goal pannot ctxt (fn _ =>
          initialise_proof_tac starting_assumption_of_proof pannot ctxt
          THEN tac))
  in
    (*NOTE we could return the theorem value alone, since
       users could get the thy value from the thm value.*)
    (thy', theorem)
  end
*}

end